""" pyduqmain.py

Overview
--------
pyduq is a data quality validation tool that implements
the LANG data quality algorithms 
(see: Zhang, R., Indulska, M., & Sadiq, S. (2019). Discovering Data Quality Problems: The Case of Repurposed Data. 
Business and Information Systems Engineering, 61(5), 575–593. https://doi.org/10.1007/s12599-019-00608-0


USAGE
--------

$ python pyduqmain.py -h

usage: pyduqmain.py [-h] [-i INPUTFILE] [-o OUTPUTFOLDER] [-m METAFILE]
                    [-s SQL SQL] [-c CUSTOM] [-p] [-v] [--infer] [--verbose]

Perform a data quality validation.

optional arguments:
  -h, --help            show this help message and exit
  -i INPUTFILE, --inputfile INPUTFILE
                        the path and name of the input data file
                        to use as the data source.
  -o OUTPUTFOLDER, --outputfolder OUTPUTFOLDER
                        the destination path for the output files to be
                        stored.
  -m METAFILE, --metafile METAFILE
                        the filename of the metadata-data file to use for
                        validation.
  -s SQL SQL, --sql SQL SQL
                        a database connection string and SQL query to use 
                        as the data source.
  -c CUSTOM, --custom CUSTOM
                        The class path and name of a custom validator.
  -p, --profile         profile the data.
  -v, --validate        validate the data.
  --infer               Generate metadata.
  --extend              Must be used with -m switch. Extends the provided
                        meatdata file with any additional attributes or
                        rules found in the data source. Overwrites the 
                        suppled metafile.
  --verbose             Generate verbose output.


OUTPUT:
--------
The outout from pylang depends on what you want to achieve.
the --profile switch will trigger a data profile the source data 
and output the results as an Excel spreadsheet.

The --validate switch will trigger a validation of the source
data and the output will be a Spreadsheet of data quality
validation errors.

The --infer swutch will force the metadata to be inferred from the
data. (Note - if -v is used and there is no metadata provided then
--infer will be set).

The --profile switch generates a profile data file.

The --extend will take an existing metadatafile, compare it with the source data, and
extend the metadata by adding any missing columns or extending attribute lengths. It
will also extend enum types where found. 


"""
#!/usr/bin/python

import sys
import os
import argparse
import pyodbc
import time
from pyduq.abstractduqvalidator import AbstractDUQValidator
from pyduq.SQLTools import SQLTools
from pyduq.duqvalidator import DUQValidator
from pyduq.patterns import Patterns
from pyduq.duqerror import ValidationError
from pyduq.filetools import FileTools
from pyduq.dataprofile import DataProfile


class pyDUQMain(object):

    def __init__(self, inputFile:str, outputFolder:str="", filePrefix:str="", metaFile:str="", customValidator:str="", sqlURI:str="", sqlQuery:str=""):
        self.metadata = {}
        self.dataset = {}
        self.inputFile = (inputFile if not inputFile is None else "sql-query")
        self.metaFile = metaFile
        self.outputFolder = (outputFolder if not outputFolder is None else ".")
        self.customValidator = customValidator
        self.sqlURI = sqlURI
        self.sqlQuery = sqlQuery
        
        if ((not filePrefix is None) and (len(filePrefix) > 0)):
            self.outputFilePrefix = filePrefix
        else:
            self.outputFilePrefix = self.inputFile.rsplit(".", 1)[0]
            if (self.outputFilePrefix.endswith("-") or self.outputFilePrefix.endswith("_")):
                self.outputFilePrefix = self.outputFilePrefix[:-1]
       
       
    def loadSQL(self):
        cnxn = pyodbc.connect(self.sqlURI)
        cursor = cnxn.cursor()
        cursor.execute(self.sqlQuery) 
        self.dataset = SQLTools(cursor).dataset
        print("SQL query returned " + str(len(self.dataset)) + " columns.")


    def loadMeta(self, extendFlag:False):
        self.metadata = FileTools.JSONtoMeta(self.metaFile)
        if(extendFlag):
            self.metadata = FileTools.extendMeta(self.metadata, FileTools.inferMeta(self.dataset))
            if (self.metaFile is None or len(self.metaFile)==0):
                self.metaFile = self.outputFilePrefix + "_meta.json"
            FileTools.MetatoJSONFile(self.metaFile, self.metadata)
            print("Using extended metadata.")


    def inferMeta(self):
        stime = time.time()
        self.metadata = FileTools.inferMeta(self.dataset)
        if (self.metaFile is None or len(self.metaFile)==0):
            self.metaFile = self.outputFilePrefix + "_meta.json"
            
        FileTools.MetatoJSONFile(self.metaFile, self.metadata)
        print("Metadata file generated in " + str(time.time() - stime) + " secs")


    def loadCSV(self):
        self.dataset = FileTools.csvFileToDict(self.inputFile)
        print("CSV file loaded " + str(len(self.dataset)) + " columns.")

    
    def loadXLS(self, sheet:None):
        self.dataset = FileTools.xlsFileToDict(self.inputFile, sheet)
        print("Excel spreadsheet loaded " + str(len(self.dataset)) + " columns.")


    def validate(self):
        try:
            stime = time.time()
     
            lang_validator = DUQValidator(self.dataset, self.metadata)
            lang_validator.validate(self.customValidator)
            
            lang_validator.saveCounters(self.outputFilePrefix + "_counters.xlsx")
            lang_validator.saveCountersSummary(self.outputFilePrefix + "_summary.xlsx")
            
            print("Validation completed in " + str(time.time() - stime) + " secs")

        except ValidationError as e:
            print (e)


    def profile(self):
        try:
            stime = time.time()
            
            data_profile = DataProfile().profile(self.metadata, self.dataset)
            FileTools.saveProfile(self.outputFilePrefix + "_profile.xlsx", data_profile)
            
            print("Profile completed in " + str(time.time() - stime) + " secs")

        except ValidationError as e:
            print (e)

    
        
def main(argv):

    # Create the parser
    my_parser = argparse.ArgumentParser(description='Perform a data quality validation.')

    # Add the arguments
    my_parser.add_argument('-i',
                           '--inputfile',
                           type=str,
                           help='the path and name of the input data file.')

    my_parser.add_argument('-o',
                           '--outputfolder',
                           type=str,
                           help='the destination path for the output files to be stored.')
    
    my_parser.add_argument('-m',
                           '--metafile',
                           type=str,
                           help='the filename of the metadata-data file to use for validation.')
                           
    my_parser.add_argument('-s',
                           '--sql',
                           nargs=2,
                           type=str,
                           help='the database connection string and SQL query')
                           
    my_parser.add_argument('-c',
                           '--custom',
                           type=str,
                           help='The class path and name of a custom validator.')

    my_parser.add_argument('-f',
                           '--fileprefix',
                           type=str,
                           help='The prefix to use for all output files.')

    my_parser.add_argument('-S',
                           '--sheet',
                           type=str,
                           help='The name of the sheet to load from an Excel spreadsheet file.')

    my_parser.add_argument('-p',
                           '--profile',
                           action="store_true",
                           help='profile the data.')

    my_parser.add_argument('-v',
                           '--validate',
                           action="store_true",
                           help='validate the data.')

    my_parser.add_argument('--infer',
                           action="store_true",
                           help='Generate metadata.')

    my_parser.add_argument('--extend',
                           action="store_true",
                           help='Extend source metadata.')

    my_parser.add_argument('--verbose',
                           action="store_true",
                           help='Generate verbose output.')



    # Execute parse_args()
    args = my_parser.parse_args()
    sqlURI = ""
    sqlQuery = ""

    if ((not args.inputfile is None) and (len(args.inputfile)>0)):
        if not os.path.isfile(args.inputfile):
            print("The input file '" + args.inputfile + "' does not exist")
            sys.exit(1)
    else:
        if (args.sql is not None):
            sqlURI = args.sql[0]
            sqlQuery = args.sql[1]
        else:
            print("You must provide either an input file OR SQL connection and query. Run pyduqmain.py -h for help.")
            sys.exit(1)            

    pyduq = pyDUQMain(args.inputfile, args.outputfolder, args.fileprefix, args.metafile, args.custom, sqlURI, sqlQuery)

    profileFlag = args.profile
    validateFlag = args.validate
    inferFlag = args.infer
    extendFlag = args.extend
    __verbose__ = args.verbose
    

    
    if (len(sqlURI) >0 ):
        pyduq.loadSQL()
    elif (len(pyduq.inputFile)>0):
        if(pyduq.inputFile.endswith(".csv")):
            pyduq.loadCSV()
        elif(pyduq.inputFile.endswith(".xlsx") or pyduq.inputFile.endswith(".xltx")):
            pyduq.loadXLS(args.sheet)
        else:
            print("Unsupported source data file type. Run pyduqmain.py -h for help.")
            sys.exit(1)            

    if (not pyduq.metaFile is None):
        if (not os.path.isfile(pyduq.metaFile)):
            print("The metadata-data file '" + pyduq.metaFile + "' does not exist")
            sys.exit(1)
        pyduq.loadMeta(extendFlag)
    else:
        if (not inferFlag):
            print("No metadata file was supplied - schema will be inferred from the dataset.")
        pyduq.inferMeta()

    if (validateFlag):
        pyduq.validate()

    if (profileFlag):
        pyduq.profile()
        
    sys.exit(0)


if  __name__ =='__main__':
    main(sys.argv[1:])



import pandas as pd
import numpy as np
import plotly

from plotly.graph_objs import Scatter,Layout,Marker,Bar



# Read data

# Dataset: Life Expectancy and per capita income (Rosling) 
class BoxPlot(object):

    def plot(self, l:dict, chartTitle:str=""):
        """
        example structure of l: 
        l = [
                {'attribute' : 'mother_id', 'error_category':'Completeness'},
                {'attribute' : 'mother_id', 'error_category':'Completeness'},
                {'attribute' : 'frog', 'error_category':'Fishing'}
                
        ]
        """        

        if (l is None or len(l) == 0):
            print ("## nothing to plot ##")
            return
        
        df = pd.DataFrame.from_dict(l)
        df.replace(-1,0,inplace=True)
        
        df = df.sort_values(by=['attribute','error_dimension'])
        
        g = df.groupby(['error_dimension','attribute'])
        
        traces = list()
        
        for title, group in g:
            m=250
            
            traces.append(
                Scatter(
                    y= [group['error_dimension'].count()],
                    x= group['error_dimension'],
                    marker= plotly.graph_objs.scatter.Marker(
                                symbol="circle",
                                size=[group['error_dimension'].count()],
                                sizeref= 2.0*m / (100.0**2.0),
                                sizemode= 'area',
                                sizemin=20,
                                colorscale='Viridis',
                                showscale= False,
                                opacity=0.5,
                                line=dict (
                                    color= 'black',
                                    width= 2
                                ),
                    ),
                    mode= 'markers',
                    showlegend= True,
                    text= "Attribute: " + "<b>" + str(title[1]) + "</b>"
                )
            )
            
        # Create chart 

        # Output will be stored as a html file. 

        # Whenever we will open output html file, one popup option will ask us about if want to save it in jpeg format. #
        # If you want basic bubble chart with only one continent just include that particular trace while providing input. 
        
        if (chartTitle != ""):
            titleStr = chartTitle
        else:
            titleStr = "<b>Chart of Data Quality Errors by Count</b>"
            
        plotly.offline.plot(
        {
            "data": traces,
            "layout": Layout(

                                title=titleStr,
                                autosize=True,
                                hovermode='closest',                                            
                                xaxis= dict(
                                    title= '<b>Data Quality Dimensions</b>',
                                    zeroline= False,
                                    gridcolor='rgb(183,183,183)',
                                    showline=True,
                                    ticklen=5,
                                    gridwidth= 2
                                ),
                                yaxis=dict(
                                    title= '<b>Count of Data Quality Errors</b>',
                                    gridcolor='rgb(183,183,183)',
                                    zeroline=False,
                                    showline=True,
                                    ticklen=5,
                                    gridwidth= 2
                                )
                            )

                },
                filename='bubble_chart.html',
                image='jpeg'
        ) 



import re

class ExpressionBuilder(object):
 
    def parseExpr(self, expr:str): 
        re.purge()
        return(re.findall(r"[^[]*\[([^]]*)\]", expr))
 
    def stripBrackets(self, str):
        result=str
        
        # find the brackets and remove them from the string
        for c in ["[","]"]:
            result=result.replace(c,"")
            
        return result
    
    def merge(self, expr, d:dict):
        s=self.stripBrackets(expr)
        for k,v in d.items():
            s=s.replace(k, str(v))
            
        return s
        
    def charPosition(string:str, c):
        pos = [] #list to store positions for each 'char' in 'string'
        for n in range(len(string)):
            if string[n] == char:
                pos.append(n)
        
        return pos

        
    def multiReplace(string, replacements, ignore_case=False):
        """
        Given a string and a dict, replaces occurrences of the dict keys found in the 
        string, with their corresponding values. The replacements will occur in "one pass", 
        i.e. there should be no clashes.
        :param str string: string to perform replacements on
        :param dict replacements: replacement dictionary {str_to_find: str_to_replace_with}
        :param bool ignore_case: whether to ignore case when looking for matches
        :rtype: str the replaced string
        """
        if (ignore_case):
            replacements = dict((pair[0].lower(), pair[1]) for pair in sorted(replacements.iteritems()))

        rep_sorted = sorted(replacements, key=lambda s: (len(s), s), reverse=True)
        rep_escaped = [re.escape(replacement) for replacement in rep_sorted]
        pattern = re.compile("|".join(rep_escaped), re.I if ignore_case else 0)
        
        return pattern.sub(lambda match: replacements[match.group(0).lower() if ignore_case else match.group(0)], string)
        
        


import typing
from enum import Enum


class DataQualityDimension(Enum):
    """ MeasurementCategory: An enumeration used to define which specific data quality dimension has been violated. The definition of these comes directly from LANG
    and the UQ DKE. 
    """
    
    COMPLETENESSMANDATORY = "Completeness of Mandatory Attributes"
    COMPLETENESSOPTIONAL = "Completeness of Optional Attributes"
    PRECISION = "Precision"
    BUSINESSRULECOMPLIANCE = "Business Rule Compliance"
    METADATACOMPLIANCE = "Metadata Compliance"
    UNIQUENESS = "Uniqueness"
    NONREDUNDANCY = "Non-redundancy"
    SEMANTICCONSISTENCY = "Semantic Consistency"
    VALUECONSISTENCY = "Value Consistency"
    FORMATCONSISTENCY = "Format Consistency"
	
	
    @staticmethod
    def names() ->list:
        names = []
        for fullname in list(DataQualityDimension):
            names.append(fullname.name)
        return (names)
        
        
class DataQualityError(object):
    """ DataQualityError: This class is used to record an instance of a discreet measurement. Every measurement has a label, an optional errorCounter and an optional value.
    """

    def __init__(self:object, attribute:str, error_dimension:DataQualityDimension=None, description:str="<Unspecified>", primary_key_value:str="<Unspecified>"):
        self.attribute = attribute
        self.error_dimension = error_dimension
        self.error_descr = description
        self.primary_key_value = primary_key_value
    
    def to_dict(self):
        values = {}
        values['attribute']= self.attribute
        values['error_dimension'] =self.error_dimension
        values['description']=self.error_descr
        values['primary_key_value']=self.primary_key_value
        return values
        


import abc
import importlib
from openpyxl import Workbook
from pyduq.dataqualityerror import DataQualityError, DataQualityDimension
from pyduq.dataprofile import DataProfile
from pyduq.duqerror import ValidationError
   
class AbstractDUQValidator(abc.ABC):
    """ AbstractDQValidator: 
    Base class that any data quality validation class should implement. It provides some basic structure for a validator.
    The constructor expects a resultset dictionary and a metadata disctionary. Classes that implement this base class will have access to
    a copy of both of those objects.
    The cobstructor will also create an empty list of errors and a empty list of measurement counters.
    """

    def __init__(self:object, dataset:dict, meta:dict):
        
        if (dataset is None):
            raise ValidationError("LANG Exception: DataSet has not been set", None)
        
        if (meta is None):
            raise ValidationError("LANG Exception: metadata has not been set", None)
        
        self.metadata = meta
        self.dataset = dataset
        self.validation_errors = []
        self.data_profile = []
            
        
    def addDataQualityError(self:object, data_quality_error:DataQualityError):
        """
        Add a new dimension.
        """
        if (data_quality_error is None):
            raise ValidationError("LANG Exception: DataQualityError has not been set", None)
        
        self.validation_errors.append(data_quality_error.to_dict())
        
    
    def profileData(self:object, meta_attribute_definition:dict, colData:dict, key:str):
        if (colData is None):
            raise ValidationError("LANG Exception: Coldata has not been set", None)

        if (meta_attribute_definition is None):
            raise ValidationError("LANG Exception: meta_attribute_definition has not been set", None)
        
        profile = DataProfile()
        profile.profileData(meta_attribute_definition, colData, key)
        profile.setPosition(len(self.data_profile)+1)
        self.data_profile.append(profile.to_dict())
        

    def saveCountersSummary(self:object, outputFile):
        workbook = Workbook(write_only=True)
        sheet = workbook.create_sheet()
        headers = ["attribute"]
        
        for name in DataQualityDimension.names():
            headers.append(name)
            #headers.append(name + " SCORE")
            
        sheet.append(headers)

        summary = self.summariseCounters()
    
        for datarow in summary.values():        
            for row in datarow:
                sheet.append(list(row.values()))
            
        workbook.save(filename=outputFile)
        workbook.close()


    def summariseCounters(self:object) ->dict:        
        # get the MeasurementCategory Enum as a list
        categories = list(DataQualityDimension)
        summary = {}
        attribute_errors = {}
 
        # Construct a list of errors per attribute and store in a dict 
        for item in self.validation_errors:
            key = item["attribute"]
            if (not key in attribute_errors):
                attribute_errors[key] = []
            
            attribute_errors[key].append(item)
        
        # now count how many times each category appears for each attribute
        #for item, data in attribute_errors.items():
        for item in self.metadata:
            summary_row = {}
            summary_row["attribute"] = item

            summary[item]=[]
            data = {}
            
            if (item in attribute_errors):
                data = attribute_errors[item]
            
            for name in categories:
                error_count=0
                
                for d in data:
                    if (d["error_dimension"] == name.value):
                        error_count+=1
                
                # for each attribute create a list of dictionaries contaning a count of each category
                summary_row[name.name] = error_count
                
                '''
                # calculate an error percentage score per error item 
                # NOTE: this has been temporarily removed to allow for easier statistical analysis in external 
                # stats tools like SPSS.  
                
                try:
                    
                    score = float(error_count / len(self.dataset[item]))
                    summary_row[name.name + " SCORE"] = round(score, 6)
                   
                except Exception as e:
                    summary_row[name.name + " SCORE"] =  0
                '''
                
            summary[item].append(summary_row)

                                
        return summary
        

    def saveCounters(self:object, outputFile):
        if (len(self.validation_errors)>0):
            workbook = Workbook(write_only=True)
            sheet = workbook.create_sheet()
            headers = list(self.validation_errors[0].keys())
            sheet.append(headers)
        
            for y in self.validation_errors:
                sheet.append(list(y.values()))
        
            workbook.save(filename=outputFile)
            workbook.close()


    def customValidator(self:object, class_path:str):        
        """
        Dynamically load a class from a string in the format '<root folder>.<module filename>.<ClassName>'
        The class must inherit AbstractDQValidator and implement the validate() method.
        """
        if (class_path is None or len(class_path)==0):
            raise ValidationError("LANG Exception: class_path has not been set", None)

        class_data = class_path.split(".")
        module_path = ".".join(class_data[:-1])
        class_str = class_data[-1]

        try:      
            module = importlib.import_module(module_path)

            # Finally, we retrieve the Class
            custom_validator = getattr(module, class_str)
        except ImportError as e:
            raise(ValidationError("Unable to load: " + class_str + " from module: " + module_path, None))
        
        if (not issubclass(custom_validator, AbstractDUQValidator)):
            raise(ValidationError("The custom validator '" + self.customValidator + "' must inherit AbstractDUQValidator.", None))
        
        obj = custom_validator(self.dataset, self.metadata)
                
        obj.validate()
        self.validation_errors.extend(obj.validation_errors)
        
        del obj
        
        
    @abc.abstractmethod
    def validate(self:object, customValidator:str=None):
        pass



import typing
import math
import statistics
import collections
import string
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import CountVectorizer
from scipy import stats
from pyduq.metautils import MetaUtils
from nltk.corpus import stopwords
stopwords = stopwords.words("english")

class DataProfile(object):
    """
    Profile: This class is used to profile a dataset and record various statistics about a set of data. The purpose of this class is to provide measurable
    values that can be used for explanation and comparison.

    """


    def __init__(self:object):
        self.attribute=""
        self.type="<Unspecified>"
        self.count=0
        self.position = 0
        self.sum = 0
        self.min_val = -1
        self.max_val = -1
        self.min_len = -1
        self.max_len = -1
        self.mean = 0
        self.variance = 0
        self.median = 0
        self.stddev = 0
        self.nullCount = 0
        self.blankCount = 0
        self.most_frequent_value = ""
        self.most_frequent_count = 0
        self.least_frequent_value = ""
        self.least_frequent_count = 0
        self.position = -1
        self.memory = 0
        self.pattern_count = 0
        self.patterns = ""
        self.default_count = 0
        self.default_value = ""
        self.negative_count = 0
        self.csim=0.0

    
    def profile(self, metadata:dict, dataset:dict) ->list:
        if (metadata is None):
            raise ValidationError("LANG Exception: meta-data has not been set", None)
        elif (dataset is None):
            raise ValidationError("LANG Exception: resultset has not been set", None)
    
        profiles = []
        count = 1
        
        for meta_attribute_key in metadata.keys():
        
            # we can't presume that the meta data attribute exists in the dataset so we
            # check first.
            if (meta_attribute_key in dataset):
                profile = DataProfile().profileData(metadata[meta_attribute_key], dataset[meta_attribute_key], meta_attribute_key)
            else:
                profile = {}
                profile["attribute"] = meta_attribute_key
            
            profile["position"] = count
            count+= 1
            profiles.append(profile)
            
        return profiles


    def profileData(self, meta_attribute_definition:dict, colData:list, key:str) ->dict:
        """
        For a given column, calculate a variety of statistics.
        """

        if (colData is None):
            raise ValidationError("LANG Exception: DataSet has not been set", None)
        
        if (meta_attribute_definition is None):
            raise ValidationError("LANG Exception: metadata has not been set", None)
        
        self.attribute = key
        self.count = len(colData)

        if (MetaUtils.exists(meta_attribute_definition, "Type")):
            self.type = meta_attribute_definition["Type"]

        if (self.type == "string"):
            cleaned = list(map(self.clean_string, colData))
            try:
                v = CountVectorizer().fit_transform(cleaned)
                vectors = v.toarray()
                self.csim = round(self.cosine_sim_vectors(vectors[0], vectors[1]), 3)
                
            except Exception as e:
                pass
        
        try:
            mode = stats.mode(colData)
            if (len(mode[0]) > 0):
                self.most_frequent_value = mode.mode[0]
                self.most_frequent_count = mode.count[0]
            
            counter = collections.Counter(colData)
            self.least_frequent_value = min(colData[::-1], key=counter.get)
            self.least_frequent_count = counter[self.least_frequent_value]
    
        except Exception as e:
            #on very large datasets, Numpy may throw an out of memory error. 
            mode = -1
            print(e)
            
        vals = []

        s=set(colData)
        s.discard("")
        self.patterns = str(sorted(s))
        self.patternCount = len(s)
        
        if (MetaUtils.exists(meta_attribute_definition, "Default")):
            self.default_value = meta_attribute_definition["Default"]
        else:
            self.default_value = "<Unspecified>"
    
        for value in colData:
            self.memory += len(value)
            
            if (len(value) == 0):
                self.blankCount += 1
            elif (value == self.default_value):
                self.default_count += 1
            elif (value == "(Null)"):
                self.nullCount += 1
                

            if (len(value) < self.min_len or self.min_len == -1):
                self.min_len = len(value)
        
            if(len(value) > self.max_len or self.max_len == -1):
                self.max_len = len(value)


            val= math.nan
                    
            try:
                if (self.type in ["int","integer"]):
                    val = int(value)
                elif (self.type in ["float","number"]):
                    val = float(value)
            except Exception as e:
                pass

     
            if (not math.isnan(val)):
                self.negative_count += (val < 0)                
                self.sum += val
                vals.append(val)

                if (val < self.min_val or self.min_val == -1):
                    self.min_val = val
            
                if(val > self.max_val or self.max_val == -1):
                    self.max_val = val
                    
                        
        if (len(vals)>0):                  
            self.mean = statistics.mean(vals)                
            self.median = statistics.median(vals)
        
        if (len(vals)>=2):
            self.stddev = statistics.stdev(vals)
            self.variance = statistics.variance(vals)
        
        return self.to_dict()
    
    def cosine_sim_vectors(self, vec1, vec2) -> float:
        vec1 = vec1.reshape(1, -1)
        vec2 = vec2.reshape(1, -1)
        
        return cosine_similarity(vec1, vec2)[0][0]
        
    
    def clean_string(self:object, text):
        text = ''.join([word for word in text if word not in string.punctuation])
        text = text.lower()
        text = ' '.join([word for word in text.split() if word not in stopwords])
        
        return text
                
    def setPosition(self, position:int):
        self.position = position
        
        
    def to_dict(self):
        values = {}
        values['attribute']= self.attribute
        values['position']= self.position
        values['type'] = self.type
        values['count'] = self.count
        values['sum'] =self.sum
        values['mean'] =self.mean
        values['median'] =self.median
        values['stddev'] =self.stddev
        values['variance'] =self.variance
        values['min_value']= self.min_val
        values['max_value']=self.max_val
        values['min_len']= self.min_len
        values['max_len']=self.max_len
        values['null_count']=self.nullCount
        values['blank_count']=self.blankCount
        values['negative_count']=self.negative_count
        values['default_count']=self.default_count
        values['default_value']=self.default_value
        values['least_frequent_value']=self.least_frequent_value
        values['least_frequent_count']=self.least_frequent_count
        values['most_frequent_value']=self.most_frequent_value
        values['most_frequent_count']=self.most_frequent_count
        values['csim']=self.csim
        values['memory_consumed_bytes']=self.memory
        values['pattern_count']=self.patternCount
        values['patterns']=self.patterns
        return values
        


import pyodbc
from prettytable import PrettyTable
from pyduq.duqerror import ValidationError

class SQLTools(object):
    """ SQLTools: This is a utility class to help manage SQL resultsets.
    Note: This could be replaced by Pandas
    """

    def __init__ (self, cursor):
        self.dataset = SQLTools.resultsetToDict(cursor)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        if (self.dataset is None):
            raise ValidationError("DataSet is NULL.", None)
        
        pt = PrettyTable()
        row = next(iter(self.dataset.values())) # grab an arbritary row so we can lift the keys
        pt.field_names = row.keys()
        for i in self.dataset:
            # pull the fields out of the resultset row and add as discreet elements to print
            pt.add_row([field for field in self.dataset[field].values()])

        return str(pt)

    @staticmethod
    def resultsetToDict(cursor) -> dict:
        """
        method resultsetToDict:
        Converts a SQL result set into a dictionary of lists keyed on the column name.
        """
        data = {}

        # Get a list of the column names returned from the query
        columns = [column[0] for column in cursor.description]
                    
        resultset = cursor.fetchall()
        colindex = 0
     
        for col in columns:
            data[col]= [("(Null)" if row[colindex] is None else str(row[colindex]).strip()) for row in resultset]            
            colindex += 1
   
        return data
    
    
    @staticmethod
    def getCol(dataset:dict, col:str) -> dict:
        """
        Slice a column out of the resultset based on col.
        The result is a dictionary of col and a list of values
        """
        return {col:SQLTools.getColValues(dataset, col)}
       

    @staticmethod
    def getColValues(dataset:dict, col:str) -> list:
        """
        Given a resultset and a column return all the values for that column as a list.
        """
        if (dataset is None):
            raise ValidationError("LANG Exception: DataSet has not been set", None)
   
        return dataset[col]


    @staticmethod
    def getColValuesAsDict(dataset:dict, *argv) -> dict:
        """
        Accepts an aribrtiary set of data columns as args and returns all the column values in a single dictionary.
        """
        if (dataset is None):
            raise ValidationError("LANG Exception: DataSet has not been set", None)

        if (argv is None):
            raise ValidationError("LANG Exception: argv has not been set", None)
        
        result = dict()
        
        for arg in argv:
            result[arg] = SQLTools.getColValues(dataset, arg)
            
        return result

    @staticmethod
    def rowCount(dataset:dict):
        """
        Return the count of rows in the resultset.
        """
        if (dataset is None):
            raise ValidationError("LANG Exception: DataSet has not been set", None)

        return len(dataset)
    




class MetaUtils(object):
    """ MetaUtils: 
    This is a helper class for processing metadata attributes. 
    At some stage we might create a metadata class, in which case
    these methods would be rolled into that class.
    """
    
    @staticmethod
    def exists(meta:dict, value:str) -> bool:
        return ( (value in meta) and (not meta[value] is None) )
    
    @staticmethod
    def isTrue(meta:dict, value:str) -> bool:
        return ( (MetaUtils.exists(meta, value)) and (meta[value]==True) )
        
    @staticmethod
    def isBlankOrNull(value:str) -> bool:
        return ( (len(value)==0) or (value == "(Null)") )

    @staticmethod
    def isAllowBlank(meta:dict) -> bool:
        return ( MetaUtils.isTrue(meta, "AllowBlank") )
     
    @staticmethod    
    def isFloat(value) -> bool:
        try:
            float(value)
            return True
        except Exception as e:
            return False    
    
    @staticmethod
    def isInt(value) -> bool:
        try:
            int(value)
            return True
        except Exception as e:
            return False



class Error(Exception):
    """Error: A generic extension of the Exception base class, used to form the basis
    of all LANG errors.
    """
    pass

class DUQError(Error):
    """ DUQError: The LANG base class for an error. All LANG errors must
    extend this class.
    
    Attributes:
        expression -- expression in which the error occurred
        message -- explanation of the error
    """
    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


class ValidationError(DUQError):
    """
    ValidationError: An implementation of LangError specificlaly for handling lang validation errors.
    """
    def __init__(self, error, message):
        self.error = error
        self.message = message




import csv
import json
import dicttoxml
from xml.dom.minidom import parseString
from unidecode import unidecode
from openpyxl import Workbook,load_workbook
from pyduq.duqerror import ValidationError

class FileTools(object):
    """ FileTools: 
    This is a utility class to help manage data files.
    The file is expected to be in a CSV format with a header row.
    Note: This could be replaced by Pandas
    """

    @staticmethod
    def csvFileToDict(fileName:str) -> dict:
        """ csvFileToDict:
        Converts a csv file into a dictionary of dictionaries.
        Each row is its own dictionary with each attribute recorded as a tupple,
        indexed by a row counter.
        """
        data = dict()
        resultset = []
        
        # first we load the data into a simple 
        with open(fileName, errors='ignore') as f:
            reader = csv.DictReader(f)
                               
            # Get a list of the column names returned from the file
            columns = list(reader.fieldnames)
            
            #for row in reader:
            #    resultset.append(row)
            resultset = [row for row in reader]
            
            for col in columns:
                data[col]= [("(Null)" if row[col] is None else FileTools.FormatString(str(row[col]).strip())) for row in resultset]            
            f.close()
            
        return data


    @staticmethod
    def xlsFileToDict(fileName:str, sheet_name:str=None) -> dict:
        """ xlsFileToDict:
        Converts an Excel spreadsheet into a dictionary of dictionaries.
        Each row is its own dictionary with each attribute recorded as a tupple,
        indexed by a row counter.
        
        Assumptions: The spreadsheet is well-formatted columns and rows.
        """
        data = {}
        
        # first we load the data into a simple
        workbook = load_workbook(filename=fileName, data_only=True, read_only = True)
        
        if (not sheet_name is None):
            if (sheet_name in workbook.sheetnames):
                sheet = workbook[sheet_name]
            else:
                raise ValidationError("Sheet '" + sheet_name + "' not found", None)
        else:        
            sheet = workbook.active
        
        # extract the column headers - assumes headers in row 1 only - perhaps 
        # this could be configrable :-)
        columns = next(sheet.iter_rows(min_row=1, max_row=1, values_only=True))

        # convert the data from rows into columns. This looks clunky
        # but it's somehow faster that iterating through the columns and
        # and provides an oportunity to clean up indiuvidual cell values.
        for row in sheet.iter_rows(min_row=2, min_col=1, values_only=True):
            col=0
            
            for value in row:                
                col_name = ("<Undefined_" + str(col) + ">" if columns[col] is None else columns[col])
                    
                if (not col_name in data):
                    data[col_name] = []
                    
                data[col_name].append("(Null)" if value is None else FileTools.FormatString(str(value).strip()))            
                col+=1
        
        workbook.close()
        return data

  
    @staticmethod
    def JSONtoMeta(fileName:str) ->dict:
        meta = {}
        
        with open(fileName, 'r', errors='ignore') as json_file:
            meta = json.load(json_file)
            json_file.close()
            
        return meta
  
    @staticmethod
    def MetatoXMLFile(xml_filename:str, meta:dict):
        xml = dicttoxml.dicttoxml(meta)
        dom = parseString(xml)

        with open(xml_filename, 'w') as w:
            w.write(dom.toprettyxml())
            w.close()
            
    @staticmethod
    def MetatoJSONFile(JSON_filename:str, meta:dict):
        json_dict = json.dumps(meta, indent=4)
        
        with open(JSON_filename, 'w') as w:
            w.write(json_dict)
            w.close()
    
    @staticmethod
    def FormatString(s:str) ->str:
        # converts a unicode string to ascii
        
        if isinstance(s, str):
            try:
                s.encode('ascii')
                return s
            except:
                return unidecode(s)
        else:
          return s


    @staticmethod
    def extendMeta(source_meta:dict, target_meta:dict) ->dict:
        """
            extendMeta: Compares a source and target metadata set
            and builds a new metadata using the combined rules.
            Attribute lengths will be whichever is longest
            Enums are merged (minus duplicates) and sorted
            Existing columns are retained intact 
            Any new columns are added
        """
        print("Extending the metadata...","\r")

        new_meta = {}
        
        for attribute_key in source_meta:
            print("Analysing the data...\t" + attribute_key)
            
            new_meta[attribute_key] = []
            new_metarow = {}

            # grab the correspoding row in the target meta. If the attribute doesn't exist then just add the source row as is 
            # to the new metadata.
            source_row = source_meta[attribute_key]
            target_row = target_meta[attribute_key]

            if (not target_row is None):
                for item_key in source_row:
                    if (item_key in target_row):
                    
                        # for attribute size we grab whichever is the biggest value
                        if (item_key == "Size"):    
                            if (int(target_row[item_key]) > int(source_row[item_key])):
                                new_metarow[item_key] = target_row[item_key]
                            else:
                                new_metarow[item_key] = source_row[item_key]
                                
                        # for enums we merge the two sets & strip duplicates
                        # note that this is a case senitive compare on purpose - 
                        # if the same value appears in different case then we want to see both
                        # so we can detemrine which (or both) are acceptable.
                        elif (item_key == "Enum"):
                            for value in target_row[item_key]:
                                if (not value in source_row[item_key]):
                                    source_row[item_key].append(value)

                            new_metarow[item_key] = sorted(source_row[item_key])
                        
                        else:
                            new_metarow[item_key] = source_row[item_key]
                    else:
                        new_metarow[item_key] = source_row[item_key]
    
                # now pick up any rules from the target row that aren't on the source
                for item_key in target_row:
                    if (not item_key in new_metarow):
                        print("\tAdding new rule '" + item_key + "' to column '" + attribute_key + "'")
                        new_metarow[item_key] = target_row[item_key]
                
            else:
                new_metarow = source_row
            
            new_meta[attribute_key] = new_metarow
        
        # finally, we merge any new attributes that exist in the target meta but not in the source
        for attribute_key in target_meta:
            if (not attribute_key in new_meta):
                print("\tAdding new column '" + attribute_key + "'")
                new_meta[attribute_key] = target_meta[attribute_key]
            
        return new_meta


    @staticmethod
    def inferMeta(dataset:dict) ->dict:
        meta = {}
    
        print("Infering the metadata...","\r")

        for attribute_key, attributes in dataset.items():
            print("Analysing the data...\t" + attribute_key)

            metarow = {}
            
            isMandatory = True
            isInt = True
            isFloat = True
            isBool = True
            isDate = False
            
            seen=set()
            size=0

                        
            for value in attributes:                
                #check to see if there are any blanks or nulls in order to determine if values are mandatory
                if (len(value)==0 or value == "(Null)"):
                    isMandatory = False
                else: 
                    #this will help determine if there are any duplicates                
                    if (not value in seen):
                        seen.add(value)

                    #calculate the size of the column
                    if (len(value)>size):
                        size = len(value)
                    
                    #test for integer
                    try:
                        int(value)
                    except Exception as e:
                        isInt = False

                    #test for float
                    try:
                        float(value)
                    except Exception as e:
                        isFloat = False
                    
                    if (not value.lower() in ["0","1","no","yes","false","true","n","y","f","t"]):
                        isBool = False
            
                    if (attribute_key.lower().find("date") != -1):
                        isDate = True
            
            metarow["Size"] = size
            metarow["Mandatory"] = isMandatory
            metarow["AllowBlank"] = not isMandatory
            metarow["Type"] = ("int" if isInt else "float" if isFloat else "bool" if isBool else "date" if isDate else "string")
            metarow["Unique"] = (len(seen) == len(attributes))

            #if we have a small number of items in our seen set then lets presume this is an enumerated attributed
            #note - there is no minimum number of set elements - an alternative approach is to capture the top n repeating strings -
            #but you're likely to get the same outcome
            if (len(seen)< 100):
                metarow["Enum"] = sorted(list(seen))
            
            meta[attribute_key]=metarow
            
        return meta
    
    @staticmethod
    def saveProfile(outputFile, data_profile:list):
        if (len(data_profile)>0):
            workbook = Workbook(write_only=True)
            sheet = workbook.create_sheet()
            c=data_profile[0]
            headers = list(c.keys())
            sheet.append(headers)
            
            for x in data_profile:
                sheet.append(list(x.values()))
        
            workbook.save(filename=outputFile)
            workbook.close()

    @staticmethod
    def listToXLS(output_file:str, l:list, headers:list):
        workbook = Workbook(write_only=True)
        sheet = workbook.create_sheet()
        sheet.append(headers)

        sheet.append(l)
            
        workbook.save(filename=output_file)
        workbook.close()
        


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyduq",
    version="1.0.5",
    author="Shane J. Downey",
    author_email="shane.downey69au@gmail.com",
    description="A tool to validate data accoridng to the University of Queensland conformed dimensions of data quality.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sjdowney/pyduq",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Freely Distributable",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "dicttoxml",        
        "plotly",
        "pyodbc",
        "prettytable",
        "openpyxl",
        "sklearn",
        "scipy",
        "nltk",
        "unidecode"
    ],
    python_requires='>=3.4',
)


import unittest

class BasicTestSuite(unittest.TestCase):

    """Basic test cases."""



    def test_absolute_truth_and_meaning(self):

        assert True





if __name__ == '__main__':

    unittest.main()


import re
import time
from pyduq.metautils import MetaUtils
from pyduq.abstractduqvalidator import AbstractDUQValidator
from pyduq.patterns import Patterns
from pyduq.duqerror import ValidationError
from pyduq.dataqualityerror import DataQualityError, DataQualityDimension
from pyduq.expressionbuilder import ExpressionBuilder
from pyduq.SQLTools import SQLTools
from pyduq.dataprofile import DataProfile

 
class DUQValidator(AbstractDUQValidator):
    """ DUQValidator: A generic validator for LANG. 
    The main execution method is validate().
    """   
        
    def validate(self:object, customValidator:str=None):
        """
        Validate a resultset against predefined metadata based on the LANG rules of data quality.
        """
        if (self.metadata is None):
            raise ValidationError("LANG Exception: meta-data has not been set", None)
        elif (self.dataset is None):
            raise ValidationError("LANG Exception: resultset has not been set", None)

        """
        Change request: find and output the primary key in the error report file if specified
        """
        primary_key = ""
        primary_key_values = None
        
        for key, item in self.metadata.items():                
            if (MetaUtils.isTrue(item, "PrimaryKey")):
                primary_key = key
                primary_key_values = self.dataset[primary_key]
                break
                
        """
        Execute a series of validations against the supplied column of data and the metadata for the column.
        Which validation is run is determined by entries in the metadata.
        """         
        for meta_attribute_key, meta_attribute_definition in self.metadata.items():                
            if (meta_attribute_key in self.dataset):
                print("Validating attribute \t'" + meta_attribute_key + "'...", end='\r')
                                
                attribute = self.dataset[meta_attribute_key]
                                
                for row_count in range(len(attribute)):
                    value = attribute[row_count]
                    
                    """ 
                    If a primarykey tag has been found then output the value so that the user 
                     has a reference to search for the record in the source system. 
                     If there is no primary key attribute set then output the row count 
                    """
                    
                    if (not primary_key_values is None):
                        primary_key_value = primary_key_values[row_count]
                    else:
                        primary_key_value = "Row: " + str(row_count+1)
                    
                    self.checkMandatory(meta_attribute_definition, meta_attribute_key, value, primary_key_value)                  
                    self.checkSize(meta_attribute_definition, meta_attribute_key, value, primary_key_value)
                    self.checkType(meta_attribute_definition, meta_attribute_key, value, primary_key_value)
                    self.checkEnum(meta_attribute_definition, meta_attribute_key, value, primary_key_value)
                    self.checkStartsWith(meta_attribute_definition, meta_attribute_key, value, primary_key_value)

                
                # format check (must provide a regex)
                if (MetaUtils.exists(meta_attribute_definition, "Format")):
                    re.purge()
                    regex=re.compile(meta_attribute_definition["Format"])
                    
                    for row_count in range(len(attribute)):
                        primary_key_value = primary_key_values[row_count]
                        value = attribute[row_count]
                        
                        isMatch = (not regex.match(value) is None)
                        
                        if ( (not isMatch) and (not MetaUtils.isAllowBlank(meta_attribute_definition)) ):
                            self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.FORMATCONSISTENCY.value, description="Error: Value '" + value + "' does not match regex #'" + meta_attribute_definition["Format"] + "'"))

                   
                # unique field check        
                if (MetaUtils.isTrue(meta_attribute_definition, "Unique") ):
                    # quick count the number of times values occurs in the column. Assumes possibly sorted so breaks the loop if >1 occurences to save time0
                    seen = set()          

                    for row_count in range(len(attribute)):
                        primary_key_value = primary_key_values[row_count]
                        value = attribute[row_count]

                        if (not value in seen):
                            seen.add(value) #only process a value once 
                        else:    
                            self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.UNIQUENESS.value, description="Error: Value '" + value + "' is not UNIQUE. A unique value was expected."))
                            
                self.checkComposite(meta_attribute_definition, meta_attribute_key)
                
                # expression evaluation is different to processing field specific validations as it could link in other columns from the resultset
                self.evaluateExpression(meta_attribute_definition, meta_attribute_key)

                print("Validating attribute \t'" + meta_attribute_key + "'...\t\t..Complete.")
            else:
                self.addDataQualityError(DataQualityError(meta_attribute_key, error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Attribute '" + meta_attribute_key + "' was not found in the dataset."))
        
        # only invoke the custom validator if one has been provoded
        if (not customValidator is None and len(customValidator) > 0):
            self.customValidator(customValidator)
        
        
    def checkMandatory(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # mandatory field check
        if (MetaUtils.isTrue(meta_attribute_definition, "Mandatory") ):
            if ( (MetaUtils.isBlankOrNull(value)) and (not MetaUtils.isAllowBlank(meta_attribute_definition)) ):
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.COMPLETENESSMANDATORY.value,description="Error: Mandatory field is BLANK or NULL. A value is required.", primary_key_value=primary_key_value))                             
        else:
            # optional field check. According to LANG optional fields shpuld contain some sort of default value
            # i.e. no field shpould ever be blank or NULL.
            if ( (MetaUtils.isBlankOrNull(value)) and (not MetaUtils.isAllowBlank(meta_attribute_definition)) ):
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.COMPLETENESSOPTIONAL.value, description="Error: Optional field is BLANK or NULL. A default value is required.", primary_key_value=primary_key_value))
                
            
    def checkComposite(self, meta_attribute_definition:dict, meta_attribute_key:str):
        # unique field check
        if (MetaUtils.exists(meta_attribute_definition, "Composite")):
            # sum the number of times value appears in the row. this is faster than using list.count(value)
            list_of_attribute_keys = meta_attribute_definition["Composite"]
            attribute_keys = '+'.join(map(str, list_of_attribute_keys))
            attribute_keys = attribute_keys.replace("%1", meta_attribute_key)
                
            # populate a dictionary of just the values that are required to create the composite meta_attribute_key
            attribute_data={}
            for col in list_of_attribute_keys:
                col = col.replace("%1", meta_attribute_key)
                attribute_data[col]=SQLTools.getColValues(self.dataset, col)
                        
            seen=set()
            rowindex=0
            # convert the dictionary of columns into a list of tuples
            fields=[dict(zip(attribute_data, col)) for col in zip(*attribute_data.values())]
            
            # check to see if there is are any duplicates in the order of attribute_keys provided
            for row in fields:
                # join the values from the columns that make up the composite meta_attribute_key to form a single value
                composite_key = '|'.join(map(str, row.values()))
                if (composite_key in seen):
                    self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.UNIQUENESS.value, description="Error: Duplicate composite meta_attribute_key: '" + attribute_keys + "', values: '" + composite_key + "'", primary_key_value=composite_key))
                else:
                    seen.add(composite_key)
               
                        
    def checkSize(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # field length check
        if (MetaUtils.exists(meta_attribute_definition, "Size")):
            if ( (len(value) > int(meta_attribute_definition["Size"])) and (not MetaUtils.isBlankOrNull(value)) ):
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' is longer than size '" + str(meta_attribute_definition["Size"]) + "'", primary_key_value=primary_key_value))
                
            
    def checkType(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # field type check
        is_valid_type = True

        if (MetaUtils.exists(meta_attribute_definition, "Type")):
            # if a default value has been specified then ignore the type check if the value matches the default
            if (MetaUtils.exists(meta_attribute_definition, "Default")):
                if (value==meta_attribute_definition["Default"]):
                    is_valid_type = False
            
            if (meta_attribute_definition["Type"] in ["int","integer"]):
                if ( (MetaUtils.isBlankOrNull(value)) or (not MetaUtils.isInt(value)) ):
                    if (not MetaUtils.isAllowBlank(meta_attribute_definition)):
                        self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' is not an int. An int was expected",primary_key_value=primary_key_value))
                        is_valid_type = False
            elif (meta_attribute_definition["Type"] in ["float","number"]):
                if ( (MetaUtils.isBlankOrNull(value)) or (not MetaUtils.isFloat(value)) ): 
                    if (not MetaUtils.isAllowBlank(meta_attribute_definition)):
                        self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' is not a float. A float was expected",primary_key_value=primary_key_value))
                        is_valid_type = False
            elif (meta_attribute_definition["Type"] in ["bool","boolean"]):
                if ( (MetaUtils.isBlankOrNull(value)) or (not value.lower() in ["false", "true", "f", "t", "n", "y", "no", "yes", "0", "1"]) ): 
                    if (not MetaUtils.isAllowBlank(meta_attribute_definition)):
                        self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' is not a boolean. A boolean was expected",primary_key_value=primary_key_value))
                        is_valid_type = False
                    
            # given that min and max checks only apply to int and float values we may as well test for them now
            if (is_valid_type):
                self.checkMinMax(meta_attribute_definition, meta_attribute_key, value, primary_key_value)
                

    def checkMinMax(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # field value range check (int and float only although in theory we could specify min and max ranges for other attributes)
        min = -1
        max = -1
        val = -1
        default = -1
        
        if (MetaUtils.exists(meta_attribute_definition, "Min")):
            try:
                min = float(meta_attribute_definition["Min"])
            except Exception as e:
                pass
        
        if (MetaUtils.exists(meta_attribute_definition, "Max")):
            try:
                max = float(meta_attribute_definition["Max"])
            except Exception as e:
                pass

        if (MetaUtils.exists(meta_attribute_definition, "Default")):
            try:
                default = float(meta_attribute_definition["Default"])
            except Exception as e:
                pass

        try:
            val = float(value)
        except Exception as e:
            pass

        if (min != -1):
            if (val != -1 and val < min and val != default):
                # error
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' must be >= " + str(min), primary_key_value=primary_key_value))
            
                
        if (max != -1):
            if (val != -1 and val > max and val != default):
                # error
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' must be <= " + str(max), primary_key_value=primary_key_value))
            

    def checkEnum(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # enumerated field check
        if (MetaUtils.exists(meta_attribute_definition, "Enum")):
            # enum is expected to be a list
            enum = meta_attribute_definition["Enum"]

            if (MetaUtils.exists(meta_attribute_definition, "Default")):
                if (value==meta_attribute_definition["Default"]):
                    # if trhe value equals the default then do nothing
                    return
           
            # check that the value exists within the provided list. If the value is blank then ignore it 
            # as we should have picked it up in the mandatory/optional test anyway
            # (i.e. if the field is optional but a value has been provided then we check it against the supplied list)
            if ( (len(value)>0) and (value not in enum) and (value != "(Null)") ):
                self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.METADATACOMPLIANCE.value, description="Error: Value '" + value + "' is outside the enumeration set '" + str(enum) + "'", primary_key_value=primary_key_value))



    def checkStartsWith(self, meta_attribute_definition:dict, meta_attribute_key:str, value:str, primary_key_value:str):
        # enumerated field check
        if (MetaUtils.exists(meta_attribute_definition, "StartsWith")):
            # startsWith is expected to be a list
            startsWith = meta_attribute_definition["StartsWith"]
            
            # check that the value exists within the provided list. If the value is blank then ignore it 
            # as we should have picked it up in the mandatory/optional test anyway
            # (i.e. if the field is optional but a value has been provided then we check it against the supplied list)
            if ( (len(value)>0) and (value != "(Null)") ):
                found = False
                for s in startsWith:
                    if value.startswith(s):
                        found = True
                        break
                        
                if (not found):
                    self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.FORMATCONSISTENCY.value, description="Error: Value '" + value + "' does not begin with any of: '" + str(startsWith) + "'", primary_key_value=primary_key_value))




            
    def evaluateExpression(self, meta_attribute_definition:dict, meta_attribute_key:str):
        # evaluate any custom expressions
        if (MetaUtils.exists(meta_attribute_definition, "Expression")):
            expr = meta_attribute_definition["Expression"]
            
            # %1 is a placeholder for whatever the column name is owning the expression (it's just a shortcut)
            expr = expr.replace("%1", "[" + meta_attribute_key + "]")
            exp = ExpressionBuilder()
            
            fields = exp.parseExpr(expr)
            colData = dict()
            
            # grab all of the columns that we need and store in a local dict
            for field in fields:
                
                # grab the column data out of the resultset
                values = SQLTools.getCol(self.dataset, field)
                
                # if the column couldn't be found then we have a configuration issue so raise an exception
                if (values is None):
                    raise ValidationError("Error evaluating expression: '" + expr + "'. Unable to find column '" + field + "' in the resultset", None)
                
                colData.update(values)
            
            # convert the seperate columns into an array of name,value pairs
            pairs=[dict(zip(colData, col)) for col in zip(*colData.values())]
            
            for pair in pairs:
                result=None
                ev = exp.merge(expr,pair)
                
                try:
                    result = eval(ev)
                except Exception as e:                    
                    self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.BUSINESSRULECOMPLIANCE.value, description="Error: Expression '" + ev + "' returned an error '" + str(e) + "'"))
                    result=None

                if ( (not result is None) and (result == False) ):
                    self.addDataQualityError(DataQualityError(meta_attribute_key,error_dimension=DataQualityDimension.BUSINESSRULECOMPLIANCE.value, description="Error: Expression '" + ev + "' returned FALSE"))

    